# **************************************************************************
# *
# * Authors:     David Herreros (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import numpy as np
import time
import os
from pyworkflow.protocol import Protocol
from pyworkflow import utils as pwutils
from pyworkflow.gui.dialog import ToolbarListDialog
from pyworkflow.gui.tree import TreeProvider
from tomo.objects import SetOfTiltSeriesCoordinates

from .viewer_triangulations import TriangulationPlot, guiThread
from .viewer_mrc import MrcPlot
from ..utils import delaunayTriangulation

from tomo.utils import extractVesicles, initDictVesicles, normalFromMatrix
import tomo.constants as const

class Tomo3DTreeProvider(TreeProvider):
    """ Populate Tree from SetOfTomograms or SetOfTiltSeries"""

    def __init__(self, tomoList):
        TreeProvider.__init__(self)
        self.tomoList = tomoList

    def getColumns(self):
        return [('Item', 300), ("# coords", 100)]

    def getObjectInfo(self, tomo):

        tomogramName = tomo.getTsId()

        return {'key': tomogramName, 'parent': None,
                'text': tomogramName,
                'values':(tomo.count),
                'tags': ("done")}

    def getObjectPreview(self, obj):
        return (None, None)

    def getObjectActions(self, obj):
        return []

    def _getObjectList(self):
        """Retrieve the object list"""
        return self.tomoList

    def getObjects(self):
        objList = self._getObjectList()
        return objList

    def configureTags(self, tree):
        tree.tag_configure("done", foreground="black")


class Tomo3DDialog(ToolbarListDialog):
    """
    This class extend from ListDialog to allow calling
    a pyvista viewer subprocess from a list of Tomograms.
    """

    def __init__(self, parent, coordinates, extnormals, **kwargs):
        self.dictVesicles, _ = initDictVesicles(coordinates)
        self.coordinates = coordinates
        self.extnormals = extnormals
        self.provider = kwargs.get("provider", None)
        ToolbarListDialog.__init__(self, parent,
                                   "Tomogram List",
                                   allowsEmptySelection=False,
                                   itemDoubleClick=self.doubleClickOnTomogram,
                                   **kwargs)

    def doubleClickOnTomogram(self, e=None):
        self.tomo = e
        # tomoName = pwutils.removeBaseExt(self.tomo.getFileName())
        tomoName = pwutils.removeBaseExt(self.tomo.get())
        self.dictVesicles = extractVesicles(self.coordinates, self.dictVesicles, tomoName)
        self.createViewer()

    def createViewer(self):
        # tomoName = pwutils.removeBaseExt(self.tomo.getFileName())
        tomoName = pwutils.removeBaseExt(self.tomo.get())
        vesicles = self.dictVesicles[tomoName]['vesicles']
        shells = []
        for vesicle in vesicles:
            shells.append(delaunayTriangulation(vesicle))

        if self.extnormals is not None:
            normals = []
            extcoords = []
            normals.append([])
            for subtomo in self.extnormals:  # separate normals by tomoName/volId
                if pwutils.removeBaseExt(subtomo.getVolName()) == tomoName:
                    normal = normalFromMatrix(subtomo.getTransform().getMatrix())
                    coord = subtomo.getCoordinate3D()
                    coordSubtomo = [coord.getX(const.SCIPION),
                                    coord.getY(const.SCIPION),
                                    coord.getZ(const.SCIPION)]
                    normals[0].append(normal)
                    extcoords.append(coordSubtomo)

        else:
            normals = self.dictVesicles[tomoName]['normals']
            extcoords = None

        classArgs = {'meshes': shells, 'clouds': vesicles, 'extNormals_List': normals, 'extNormals_coords': extcoords}
        guiThread(TriangulationPlot, 'initializePlot', **classArgs)

class ViewerMRCDialog(ToolbarListDialog):
    """
    This class extend from ListDialog to allow calling
    a pyvista viewer subprocess from a list of Tomograms.
    """

    def __init__(self, parent, coords, protocol:Protocol, **kwargs):
        self.coords = coords
        self.prot = protocol
        self.provider = kwargs.get("provider", None)
        ToolbarListDialog.__init__(self, parent,
                                   "Tomogram list",
                                   allowsEmptySelection=False,
                                   itemDoubleClick=self.doubleClickOnTomogram,
                                   **kwargs)

    def doubleClickOnTomogram(self, tomo=None):
        tomo_path = tomo.getFileName()
        coord_list = []
        direction_list = []
        boxSize = 32

        def fromCoordinated3D():
            for coord in self.coords.iterCoordinates(volume=tomo):
                direction = normalFromMatrix(coord.getMatrix())
                position = [coord.getX(const.BOTTOM_LEFT_CORNER),
                            coord.getY(const.BOTTOM_LEFT_CORNER),
                            coord.getZ(const.BOTTOM_LEFT_CORNER),
                            coord.getObjId()]
                position.append(coord.getGroupId()) if coord.getGroupId() is not None else position.append(0)
                coord_list.append(position)
                direction_list.append(direction)
            boxSize = self.coords.getBoxSize()

        def fromTiltSeriesCoordinates():
            for coord in self.coords.iterItems(where="_tsId='%s'" % tomo.getTsId()):
                direction = normalFromMatrix(np.eye(4))
                position = [coord.getX(),
                            coord.getY(),
                            coord.getZ(),
                            coord.getObjId(),
                            0] # Group Id
                coord_list.append(position)
                direction_list.append(direction)

        if not isinstance(self.coords, SetOfTiltSeriesCoordinates):
            fromCoordinated3D()
        else:
            fromTiltSeriesCoordinates()

        posFile = self.prot.getPath('positions.txt')
        directionsFile= self.prot.getPath('directions.txt')
        np.savetxt(posFile, np.asarray(coord_list))
        np.savetxt(directionsFile, np.asarray(direction_list))
        viewer_args = {'tomo_mrc': tomo_path, 'points': posFile, 'normals': directionsFile,
                       'boxSize': boxSize}

        guiThread(MrcPlot, 'initializePlot', **viewer_args)

    def haveCoordinatesChanged(self, coordinateFiles):
        hasChanged = False
        if len(coordinateFiles):
            for coordFile, oldCreationTime in coordinateFiles.items():
                if os.path.isfile(coordFile):
                    newCreationTime = time.ctime(os.path.getctime(coordFile))
                    if newCreationTime != oldCreationTime:
                        hasChanged = True
                        break
        return hasChanged


# **************************************************************************
# *
# * Authors:     David Herreros Calero (dherreros@cnb.csic.es)
# *
# * Unidad de  Bioinformatica of Centro Nacional de Biotecnologia , CSIC
# *
# * This program is free software; you can redistribute it and/or modify
# * it under the terms of the GNU General Public License as published by
# * the Free Software Foundation; either version 2 of the License, or
# * (at your option) any later version.
# *
# * This program is distributed in the hope that it will be useful,
# * but WITHOUT ANY WARRANTY; without even the implied warranty of
# * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# * GNU General Public License for more details.
# *
# * You should have received a copy of the GNU General Public License
# * along with this program; if not, write to the Free Software
# * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
# * 02111-1307  USA
# *
# *  All comments concerning this program package may be sent to the
# *  e-mail address 'scipion@cnb.csic.es'
# *
# **************************************************************************
import os.path
import time
import numpy as np
import pyworkflow.viewer as pwviewer
from pyworkflow.gui.dialog import askYesNo, showInfo
from pyworkflow.utils.properties import Message
import pyworkflow.utils as pwutils

from pwem.protocols import EMProtocol
import pwem.viewers.views as vi
from .views_tkinter_tree import Tomo3DTreeProvider
from .views_tkinter_tree import ViewerMRCDialog

import tomo.objects
from ..protocols import XmippProtFilterbyNormal


class TomoVizDataViewer(pwviewer.Viewer):
    """ Wrapper to visualize different type of objects
    using pyvista
    """
    _environments = [pwviewer.DESKTOP_TKINTER]
    _targets = [
        tomo.objects.SetOfCoordinates3D,
        XmippProtFilterbyNormal,
        tomo.objects.SetOfTiltSeriesCoordinates,
        tomo.objects.SetOfSubTomograms
    ]

    def __init__(self, **kwargs):
        pwviewer.Viewer.__init__(self, **kwargs)
        self._views = []

    def _getObjView(self, obj, fn, viewParams={}):
        return vi.ObjectView(
            self._project, obj.strId(), fn, viewParams=viewParams)

    def _visualize(self, obj, **kwargs):
        views = []
        cls = type(obj)

        itemField = tomo.objects.Tomogram.TS_ID_FIELD

        if issubclass(cls, tomo.objects.SetOfCoordinates3D):
            outputCoords = obj
            tomos = outputCoords.getPrecedents()
            groupAttribute = tomo.objects.Coordinate3D.TOMO_ID_ATTR

        elif issubclass(cls, tomo.objects.SetOfSubTomograms):
            outputCoords = obj.getCoordinates3D()
            if outputCoords is None:
                showInfo("3D coordinates missing",
                         "This set of subtomograms is not associated with a set of coordinates 3d and cannot be used in this coordinate viewer.",
                         parent=self.getParent().root)
                return []
            tomos = outputCoords.getPrecedents()

            # Keep subtomos as the main collection, since Subtomograms subsets do not affect the coordinates set
            outputCoords = obj
            groupAttribute = tomo.objects.SubTomogram.VOL_NAME_FIELD

        elif issubclass(cls, EMProtocol):
            outputCoords = obj.inputMeshes.get()
            tomos = outputCoords.getPrecedents()
            groupAttribute = "_volId"

        elif issubclass(cls, tomo.objects.SetOfTiltSeriesCoordinates):
            outputCoords = obj
            tomos = outputCoords.getSetOfTiltSeries()
            groupAttribute = "_tsId"
            itemField = groupAttribute

        volIds = outputCoords.aggregate(["MAX", "COUNT"], groupAttribute, [groupAttribute])
        volIds = [(d[groupAttribute], d["COUNT"]) for d in volIds]

        tomoList = []
        coordinateFiles = dict()

        for objId in volIds:
            tomogram = tomos[{itemField:objId[0]}].clone()
            tomogram.count = objId[1]
            tomoList.append(tomogram)
            # Check if _indices.txt files exists for the tomogram and get the
            # creation time
            basename = pwutils.removeBaseExt(tomogram.getFileName())
            indices_file = basename + '_indices.txt'
            if os.path.isfile(indices_file):
                coordinateFiles[indices_file] = time.ctime(
                    os.path.getctime(indices_file))

                # tomoList = [String(tomos[objId].getFileName()) for objId in volIds]
        tomoProvider = Tomo3DTreeProvider(tomoList)
        viewer = ViewerMRCDialog(self._tkRoot, outputCoords, self.protocol,
                                 provider=tomoProvider,
                                 allowSelect= False,
                                 cancelButton=True)

        # For now we only generate 3d coordinates
        if viewer.haveCoordinatesChanged(coordinateFiles) and issubclass(cls, tomo.objects.SetOfCoordinates3D):
            import tkinter as tk
            frame = tk.Frame()
            if askYesNo(Message.TITLE_SAVE_OUTPUT, Message.LABEL_SAVE_OUTPUT, frame):
                protocol = self.protocol
                suffix = protocol._getOutputSuffix(tomo.objects.SetOfCoordinates3D)
                updated_set = protocol._createSetOfCoordinates3D(tomos, suffix)
                updated_set.setName("Selected Coordinates")
                updated_set.setPrecedents(tomos)
                updated_set.setSamplingRate(tomos.getSamplingRate())
                updated_set.setBoxSize(outputCoords.getBoxSize())
                for item in tomoList:
                    basename = pwutils.removeBaseExt(item.getFileName())
                    indices_file = basename + '_indices.txt'
                    if os.path.isfile(indices_file):
                        indices = np.loadtxt(indices_file, delimiter=' ')
                        for index in indices:
                            updated_set.append(outputCoords[index].clone())
                        pwutils.cleanPath(indices_file)
                name = protocol.OUTPUT_PREFIX + suffix
                args = {}
                args[name] = updated_set
                protocol._defineOutputs(**args)
                protocol._defineSourceRelation(tomos, updated_set)
                protocol._updateOutputSet(name, updated_set, state=updated_set.STREAM_CLOSED)

        # TODO: This should return a list of views and not already launching the dialog.
        return []
